package org.apfloat.internal;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.HashSet;

import org.apfloat.*;

import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * @since 1.1
 * @version 1.5.1
 * @author Mikko Tommila
 */

public class ParallelRunnerTest
    extends TestCase
{
    public ParallelRunnerTest(String methodName)
    {
        super(methodName);
    }

    public static void main(String[] args)
    {
        junit.textui.TestRunner.run(suite());
    }

    public static TestSuite suite()
    {
        TestSuite suite = new TestSuite();

        suite.addTest(new ParallelRunnerTest("testRunParallel"));
        suite.addTest(new ParallelRunnerTest("testRunParallelLocked"));
        suite.addTest(new ParallelRunnerTest("testRunConcurrentTask"));
        suite.addTest(new ParallelRunnerTest("testRunConcurrentTaskRunning"));
        suite.addTest(new ParallelRunnerTest("testRunConcurrentTasks"));
        suite.addTest(new ParallelRunnerTest("testRunConcurrentTasksMultiple"));

        return suite;
    }

    public static void testRunParallel()
    {
        for (int threads = 1; threads <= 32; threads++)
        {
            final int LENGTH = 1000000;
            final int[] values = new int[LENGTH];
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            for (int i = start; i < start + length; i++)
                            {
                                values[i] += i;
                            }
                        }
                    };
                }
            };

            ParallelRunner parallelRunner = new ParallelRunner(threads);
            parallelRunner.runParallel(parallelRunnable);

            for (int i = 0; i < LENGTH; i++)
            {
                assertEquals(threads + " threads, element " + i, i, values[i]);
            }
        }
    }

    public static void testRunParallelLocked()
    {
        for (int threads = 1; threads <= 32; threads++)
        {
            final int LENGTH = 1000000;
            final int[] values = new int[LENGTH];
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            for (int i = start; i < start + length; i++)
                            {
                                values[i] += i;
                            }
                        }
                    };
                }
            };

            Object key = new Object();
            ParallelRunner parallelRunner = new ParallelRunner(threads);
            parallelRunner.lock(key);
            try
            {
                parallelRunner.runParallel(parallelRunnable);
            }
            finally
            {
                parallelRunner.unlock();
            }

            for (int i = 0; i < LENGTH; i++)
            {
                assertEquals(threads + " threads, element " + i, i, values[i]);
            }
        }
    }

    public static void testRunConcurrentTask()
    {
        // Other thread joins, task not running
        for (int threads = 2; threads <= 32; threads++)
        {
            final int LENGTH = 10000;
            final Set<String> threadNames = Collections.synchronizedSet(new HashSet<String>());
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };
            final ParallelRunnable dummyRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return 0;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                        }
                    };
                }
            };

            final Object key = new Object();

            final int otherThreads = (threads + 1) / 2;
            Thread otherThread = new Thread()
            {
                public void run()
                {
                    ParallelRunner parallelRunner = new ParallelRunner(otherThreads);
                    parallelRunner.lock(key);
                    try
                    {
                        parallelRunner.runParallel(dummyRunnable);
                    }
                    finally
                    {
                        parallelRunner.unlock();
                    }
                }
            };

            ParallelRunner parallelRunner = new ParallelRunner(threads / 2);
            parallelRunner.lock(key);
            try
            {
                otherThread.start();
                sleepUninterrupted(50);
                parallelRunner.runParallel(parallelRunnable);
            }
            finally
            {
                parallelRunner.unlock();
            }

            assertEquals(threads + " threads (" + threadNames + ")", threads, threadNames.size());
        }
    }

    public static void testRunConcurrentTaskRunning()
    {
        // Other thread joins, task running
        for (int threads = 2; threads <= 32; threads++)
        {
            final int LENGTH = 10000;
            final Set<String> threadNames = Collections.synchronizedSet(new HashSet<String>());
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };
            final ParallelRunnable dummyRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return 0;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                        }
                    };
                }
            };

            final Object key = new Object();

            final int otherThreads = (threads + 1) / 2;
            Thread otherThread = new Thread()
            {
                public void run()
                {
                    sleepUninterrupted(50);
                    ParallelRunner parallelRunner = new ParallelRunner(otherThreads);
                    parallelRunner.lock(key);
                    try
                    {
                        parallelRunner.runParallel(dummyRunnable);
                    }
                    finally
                    {
                        parallelRunner.unlock();
                    }
                }
            };

            ParallelRunner parallelRunner = new ParallelRunner(threads / 2);
            parallelRunner.lock(key);
            try
            {
                otherThread.start();
                parallelRunner.runParallel(parallelRunnable);
            }
            finally
            {
                parallelRunner.unlock();
            }

            assertEquals(threads + " threads (" + threadNames + ")", threads, threadNames.size());
        }
    }

    public static void testRunConcurrentTasks()
    {
        // Run two tasks between lock and unlock
        for (int threads = 2; threads <= 32; threads++)
        {
            final int LENGTH = 10000;
            final Set<String> threadNames = Collections.synchronizedSet(new HashSet<String>());
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };
            final boolean[] dummyDone = new boolean[1];
            final ParallelRunnable dummyRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return 1;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            dummyDone[0] = true;
                        }
                    };
                }
            };

            final Object key = new Object();

            final int otherThreads = (threads + 1) / 2;
            Thread otherThread = new Thread()
            {
                public void run()
                {
                    sleepUninterrupted(50);
                    ParallelRunner parallelRunner = new ParallelRunner(otherThreads);
                    parallelRunner.lock(key);
                    try
                    {
                        parallelRunner.runParallel(dummyRunnable);
                    }
                    finally
                    {
                        parallelRunner.unlock();
                    }
                }
            };

            Set<String> threadNames2;
            ParallelRunner parallelRunner = new ParallelRunner(threads / 2);
            parallelRunner.lock(key);
            try
            {
                otherThread.start();
                parallelRunner.runParallel(parallelRunnable);

                threadNames2 = new HashSet<String>(threadNames);
                threadNames.clear();
                parallelRunner.runParallel(parallelRunnable);
            }
            finally
            {
                parallelRunner.unlock();
            }

            assertEquals(threads + " threads (" + threadNames + ")", threads, threadNames.size());
            assertEquals(threads + " threads2 (" + threadNames2 + ")", threads, threadNames2.size());

            sleepUninterrupted(100);    // Wait for otherThread to finish
            assertTrue(threads + " threads other thread done", dummyDone[0]);
        }
    }

    public static void testRunConcurrentTasksMultiple()
        throws InterruptedException
    {
        // Three threads compete for same key; one of them contributes threads to both other threads; while another task (for another key) starts and ends
        for (int threads = 3; threads <= 32; threads++)
        {
            final int LENGTH = 10000;
            final Set<String> threadNames1 = Collections.synchronizedSet(new HashSet<String>());
            final Set<String> threadNames2 = Collections.synchronizedSet(new HashSet<String>());
            final Set<String> threadNames3 = Collections.synchronizedSet(new HashSet<String>());
            ParallelRunnable parallelRunnable = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames1.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };
            final ParallelRunnable parallelRunnable2 = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames2.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };
            final ParallelRunnable parallelRunnable3 = new ParallelRunnable()
            {
                public int getLength()
                {
                    return LENGTH;
                }

                public Runnable getRunnable(final int start, final int length)
                {
                    return new Runnable()
                    {
                        public void run()
                        {
                            threadNames3.add(Thread.currentThread().getName());
                            sleepUninterrupted(10);
                        }
                    };
                }
            };

            final Object key = new Object();

            final int otherThreads = (threads + 2) / 3;
            Thread otherThread = new Thread()
            {
                public void run()
                {
                    sleepUninterrupted(50);
                    ParallelRunner parallelRunner = new ParallelRunner(otherThreads);
                    parallelRunner.lock(key);
                    try
                    {
                        parallelRunner.runParallel(parallelRunnable2);
                    }
                    finally
                    {
                        parallelRunner.unlock();
                    }
                }
            };

            final int otherThreads2 = (threads + 1) / 3;
            Thread otherThread2 = new Thread()
            {
                public void run()
                {
                    ParallelRunner parallelRunner = new ParallelRunner(otherThreads2);
                    parallelRunner.lock(key);
                    try
                    {
                        parallelRunner.runParallel(parallelRunnable3);
                    }
                    finally
                    {
                        parallelRunner.unlock();
                    }
                }
            };

            ParallelRunner parallelRunner = new ParallelRunner(threads / 3);
            parallelRunner.lock(key);
            try
            {
                otherThread.start();
                otherThread2.start();
                parallelRunner.runParallel(parallelRunnable);
            }
            finally
            {
                parallelRunner.unlock();
            }

            assertEquals(threads + " threads (" + threadNames1 + ")", threads, threadNames1.size());

            otherThread.join();
            otherThread2.join();
            List<Set<String>> threadNames = Arrays.asList(threadNames2, threadNames3);
            if (threadNames2.size() < threadNames3.size())
            {
                Collections.reverse(threadNames);
            }

            assertEquals(threads + " threads2 (" + threadNames.get(0) + ")", otherThreads + otherThreads2, threadNames.get(0).size());
            assertTrue(threads + " threads3 (" + threadNames.get(1) + ")", Arrays.asList(otherThreads, otherThreads2).contains(threadNames.get(1).size()));
        }
    }

    private static void sleepUninterrupted(long time)
    {
        try
        {
            Thread.sleep(time);
        }
        catch (InterruptedException ie)
        {
            fail(ie.toString());
        }
    }
}
